#!/usr/bin/python

# For many installations the shebang line is:
#!/usr/bin/python

# reportsegmenter-standardio.py
# D. Gibbon

#-------------------------------------------------------------
# Standard IO version
# 2008-10-28

# CGI input versions
# 2008-10-24
# 2008-09-06
#-------------------------------------------------------------
# Method:
# 1. A functional/imperative programming style is used (not OO).
# 2. Elementary regexp filter/transformation techniques are used:
#    - Reason - transparent: BARDML is describable with regular expressions.
#    - A couple of minor changes to BARDML are introduced.
#    - Thus XML is not processed with dedicated XML handlers.
#    - But dedicated XML handlers can be introdued later if required.
# 3. Input:
#    - read BARDML file from standard in
#    - reporting-item-collection elements are ignored
#    - incorrectly, a particular ordering of XML elements is currently assumed
#    - information is internally stored in a Python dictionary (hash)
# 4. XML tokenisation:
#    - create a single long string for regexp processing
#    - mark element beginnings with '#' (may need to be changed)
#    - split elements on '#'
#    - put information into a Python dictionary (hash table):
#        - extract 'narrative' and 'titlephrase' as strings
#        - extract attribute-value pairs
# 4. Factorisation
#    - Sentences are segmented using punctuation and conjunctions.
#    - Oversegmentation occurs because of ambiguity of these items.
#    - Elementary regexp techniques provide partial solutions.
#    - More sophisticated techniques can be introduced in later versions.
# 5. Output:
#    - Format BARDML using information from dictionary
#    - write BARDML file to standard out.

#-------------------------------------------------------------
# Import Python modules
#-------------------------------------------------------------

# System (for standardio), inspect (for debugging), re (for regexp)

import sys
import inspect
import re

#-------------------------------------------------------------
# List conjunctions and (conjunctive) relative pronouns
#-------------------------------------------------------------

def conjunctions():
    return [
'and', 'but', 'or', 'nor', 'even though', 'although', 'though', 'whereas', 'however', 'thus', 'thereby', 'nevertheless', 'in the process', 'in so doing', 'in case that', 'in case', 'then', 'even if', 'only if',  'if', 'unless', 'whether or not', 'whether', 'since', 'until', 'till', 'because', 'now that', 'as', 'in order that', 'in order to', 'so that', 'so', 'before', 'after', 'while', 'where', 'when', 'who', 'whom', 'how', 'why', 'what', 'which'
]

#-------------------------------------------------------------
# Sentence tokenisation
#-------------------------------------------------------------

def factorise(narrative):

# Convert tabs and EOL to spaces
    narrative = re.sub('\t',' ',narrative)
    narrative = re.sub('\r',' ',narrative)
    narrative = re.sub('\n',' ',narrative)
    narrative = re.sub('  *',' ',narrative)

# Normalise token boundaries to spaces:
    narrative = re.sub('\. *$','',narrative)
    narrative = ' ' + narrative + ' '
    narrative = re.sub('  *',' ',narrative)

# *** To do ***
# Note: abbreviation-final period incorrectly handled

# Protect punctuation periods and commas
    narrative = re.sub('\. ','#PERIOD#',narrative)
    narrative = re.sub(', ','#COMMA#',narrative)

# Protect abbreviation and numerical periods and commas
    narrative = re.sub('\.','#PERI#',narrative)
    narrative = re.sub(',','#COMM#',narrative)

# Revert to punctuation periods and commas
    narrative = re.sub('#PERIOD#','.',narrative)
    narrative = re.sub('#COMMA#',',',narrative)

# Normalise all sentence-final points to period
    narrative = re.sub('[,:;\?\!]','.' + '--',narrative)

# Normalise all conjunctions to period (beware of empty factors!!!)
    for conjunction in conjunctions():
#        narrative = re.sub(' '+conjunction+' ', '.', narrative) 
        narrative = re.sub(' '+conjunction+' ', '. +' + conjunction + ' ', narrative) 


# Tokenise all periods
    narrative = re.sub('\.\.*',' . ',narrative)

# Normalise all spaces to one space
    narrative = re.sub('  *',' ',narrative)

# Split narrative string into list of factor strings
    factors = narrative.split(' . ')

# Revert to abbreviation and numerical periods and commas
    newfactors = []
    for string in factors:
        string = re.sub('#PERI#','.',string)
        string = re.sub('#COMM#',',',string)
        string = re.sub(' *$','.',string)
        string = re.sub('^ *','',string)
        newfactors.append(string)

    return newfactors

#-------------------------------------------------------------
# XML input tokenisation
#-------------------------------------------------------------

def xmltokenise(xml):

# Convert tabs and EOL to spaces
    xml = re.sub('\t',' ',xml)
    xml = re.sub('\r',' ',xml)
    xml = re.sub('\n',' ',xml)
# Also remove &#xD from Reporter input. DG 2008-11-12
    xml = re.sub('&#xD;',' ',xml)
    xml = re.sub('  *',' ',xml)

# Hard-wired: needs to be changed!!!
    xml = re.sub('<reporter','#<reporter',xml)
    xml = re.sub('<incident','#<incident',xml)
    xml = re.sub('<narrative','#<narrative',xml)
    xml = re.sub('<titlephrase','#<titlephrase',xml)
# These elements are currently ignored but could be used for post-editing:
    xml = re.sub('<reporting-item-collection','#<reporting-item-collection',xml)
    xml = re.sub('<reporting-item','#<reporting-item',xml)

    xml = re.sub('<\?xml version[^>]*>','',xml)
    xml = re.sub('\"','',xml)

    elementlist = xml.split('#')
    itemdict={}
    for element in elementlist:
        element = re.sub('<\/.*','',element)
        element = re.sub('\/?>',' ',element)
        element = re.sub('<',' ',element)
        itemlist = element.split()
        if len(itemlist) > 1:
            itemdict[itemlist[0]] = itemlist[1:]

# Process attribute-value pairs for 'incident' and 'reporter' elements 
# Hard-wired: needs to be changed!!!
    incident = itemdict['incident']
    reporter = itemdict['reporter']
    incidentreporter = incident + reporter

    for avp in incidentreporter:
        avplist = avp.split('=')
        if len(avplist) > 1:
            itemdict[avplist[0]] = avplist[1]
    return itemdict

#-------------------------------------------------------------
# Format XML output
#-------------------------------------------------------------

def xmlformat(itemdictionary, factors):

    narrative = itemdictionary['narrative']
    narrative = ' '.join(narrative)

    titlephrase = itemdictionary['titlephrase']
    titlephrase = ' '.join(titlephrase)

    mrmrs = ' mrmrs="' + itemdictionary['mrmrs'] + '"'
    firstName = ' firstName="' + itemdictionary['firstName'] + '"'
    initialName = ' initialName="' + itemdictionary['initialName'] + '"'
    lastName = ' lastName="' + itemdictionary['lastName'] + '"'
    institution = ' institution="' + itemdictionary['institution'] + '"'
    contact = ' contact="' + itemdictionary['contact'] + '"'
    involvement = ' involvement="' + itemdictionary['involvement'] + '"'
    role = ' role="' + itemdictionary['role'] + '"'
    persondamage = ' persondamage="' + itemdictionary['persondamage'] + '"'
    objectdamage = ' objectdamage="' + itemdictionary['objectdamage'] + '"'

# Convert tabs and EOL to spaces

    xmlhead = '<?xml version="1.0" encoding="UTF-8"?>\n\n'

    narrative = '    <narrative>' + narrative + '</narrative>\n\n'

    titlephrase = '    <titlephrase>' + titlephrase + '</titlephrase>\n\n'

    reporter = '    <reporter' + mrmrs + firstName + initialName + lastName + institution + contact + involvement + role + '/>\n\n'

    reportingitems = ''
    for statement in factors:
        reportingitems = reportingitems + "        <reporting-item type=\"ITEM\">" + statement + "</reporting-item>\n"

    reportingitemcollectionhead = '    <reporting-item-collection>\n'
    reportingitemcollectiontail = '    </reporting-item-collection>\n\n' 
    reportingitemcollection = reportingitemcollectionhead + reportingitems + reportingitemcollectiontail

    incidenthead = '<incident' + persondamage + objectdamage +  '>\n\n'
    incidenttail = '</incident>\n'
    incident = incidenthead + titlephrase + narrative + reporter + reportingitemcollection + incidenttail 

    xml = xmlhead + incident

    return xml

#-------------------------------------------------------------
# XML mime type definition
#-------------------------------------------------------------

def xmlinit():

    return "Content-type: text/xml\n\n"

#-------------------------------------------------------------
# Main caller
#-------------------------------------------------------------

def main():

    narrative = sys.stdin.read()

#    itemdictionary = xmltokenise(xmlinput)

#    narrative = itemdictionary['narrative']
#    narrative = ' '.join(narrative)

    factors = factorise(narrative)

#    xmloutput = xmlformat(itemdictionary, factors)
    for statement in factors:
        print statement



#-------------------------------------------------------------
# Start
#-------------------------------------------------------------

main()

# EOF --------------------------------------------------------

